#!/usr/bin/env python
# vim:sw=4 ts=4 et


def issub(sub, other):
    for x in sub:
        if x not in other:
            return False
    return True

class attr_view(object):
    def __init__(self, dir):
        self.dir = dir

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, list(self).__repr__())

    def __len__(self):
        return len(self.dir)

class attr_setview(attr_view):
    def __eq__(self, other):
        return len(self) == len(other) and issub(other, self)

    def __ne__(self, other):
        return not __eq__(self, other)

    def __lt__(self, other):
        return len(self) < len(other) and issub(self, other)

    def __le__(self, other):
        return len(self) <= len(other) and issub(self, other)

    def __gt__(self, other):
        return len(self) > len(other) and issub(other, self)

    def __ge__(self, other):
        return len(self) >= len(other) and issub(other, self)

    def __or__(self, other):
        return set(self) | other

    def __and__(self, other):
        return set(self) & other

    def __sub__(self, other):
        return set(self) - other

    def __xor__(self, other):
        return set(self) ^ other

class attr_viewkeys(attr_setview):
    def __contains__(self, other):
        return other in self.dir

    def __iter__(self):
        return self.dir.iterkeys()

class attr_viewvalues(attr_view):
    def __iter__(self):
        return self.dir.itervalues()

class attr_viewitems(attr_setview):
    def __contains__(self, other):
        (key, val) = other
        return key in self.dir and self.dir[key] == val

    def __iter__(self):
        return self.dir.iteritems()

class attr_viewdict(attr_viewkeys):
    """
Provide a dict-like view on an attribute directory. Nested attribute
directories are shown as dictionaries, so the view provides a deep copy
the attribute (sub-)hierarchy.
    """
    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, dict(self).__repr__())

    def keys(self):
        return list(self)

    def __getitem__(self, key):
        #this import here is needed due to circular dependency
        #with _kdumpfile
        from _kdumpfile import attr_dir
        val = self.dir[key]
        if isinstance(val, attr_dir):
            return val.copy()
        else:
            return val



#!/usr/bin/env python
# vim:sw=4 ts=4 et

from _addrxlat import *
from addrxlat.exceptions import *
import sys

class FullAddress(FullAddress):
    def __init__(self, addrspace=NOADDR, addr=0, *args, **kwargs):
        super(FullAddress, self).__init__(*args, **kwargs)
        self.addrspace = addrspace
        self.addr = addr

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.addrspace, self.addr)

    def conv(self, addrspace, ctx, sys):
        status = super(FullAddress, self).conv(addrspace, ctx, sys)
        if status != OK:
            raise get_exception(status, ctx.get_err())

    def copy(self):
        "make a copy of self"
        return type(self)(addrspace=self.addrspace, addr=self.addr)

class Context(Context):
    def __init__(self, *args, **kwargs):
        super(Context, self).__init__(*args, **kwargs)
        self.convert = convert

    def __repr__(self):
        return '%s()' % (self.__class__.__name__,)

class Method(Method):
    def __init__(self, kind, target_as=NOADDR, param=(), *args, **kwargs):
        super(Method, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as
        self.param = param

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
            self.__class__.__name__,
            self.kind,
            self.target_as,
            self.param)

class CustomMethod(CustomMethod):
    def __init__(self, target_as=NOADDR, *args, **kwargs):
        super(CustomMethod, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            self.target_as)

class LinearMethod(LinearMethod):
    def __init__(self, target_as=NOADDR, off=0, *args, **kwargs):
        super(LinearMethod, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as
        self.off = off

    def __repr__(self):
        return '%s(%r, %r)' % (
            self.__class__.__name__,
            self.target_as,
            self.off)

class PageTableMethod(PageTableMethod):
    def __init__(self, target_as=NOADDR, root=None, pte_format=PTE_NONE, fields=(), *args, **kwargs):
        super(PageTableMethod, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as
        self.root = root
        self.pte_format = pte_format
        self.fields = fields

    def __repr__(self):
        return '%s(%r, %r, %r, %r)' % (
            self.__class__.__name__,
            self.target_as,
            self.root,
            self.pte_format,
            self.fields)

class LookupMethod(LookupMethod):
    def __init__(self, target_as=NOADDR, endoff=0, tbl=(), *args, **kwargs):
        super(LookupMethod, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as
        self.endoff = endoff
        self.tbl = tbl

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
            self.__class__.__name__,
            self.target_as,
            self.endoff,
            self.tbl)

class MemoryArrayMethod(MemoryArrayMethod):
    def __init__(self, target_as=NOADDR, base=None, shift=0, elemsz=0, valsz=0, *args, **kwargs):
        super(MemoryArrayMethod, self).__init__(*args, **kwargs)
        self.convert = convert
        self.target_as = target_as
        self.base = base
        self.shift = shift
        self.elemsz = elemsz
        self.valsz = valsz

    def __repr__(self):
        return '%s(%r, %r, %r, %r, %r)' % (
            self.__class__.__name__,
            self.target_as,
            self.base,
            self.shift,
            self.elemsz,
            self.valsz)

class Range(Range):
    def __init__(self, endoff=0, meth=SYS_METH_NONE, *args, **kwargs):
        super(Range, self).__init__(*args, **kwargs)
        self.endoff = endoff
        self.meth = meth

    def __repr__(self):
        return '%s(%r, %r)' % (
            self.__class__.__name__,
            self.endoff,
            self.meth)

class Map(Map):
    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        self.convert = convert

    def __repr__(self):
        return '%s()' % (self.__class__.__name__)

    def set(self, addr, range):
        status = super(Map, self).set(addr, range)
        if status != OK:
            raise get_exception(status)

class System(System):
    def __init__(self, *args, **kwargs):
        super(System, self).__init__(*args, **kwargs)
        self.convert = convert

    def __repr__(self):
        return '%s()' % (self.__class__.__name__)

    def os_init(self, ctx, arch, *args, **kwargs):
        status = super(System, self).os_init(ctx, arch, *args, **kwargs)
        if status != OK:
            raise get_exception(status, ctx.get_err())

class Step(Step):
    def __init__(self, ctx, sys=None, meth=None, *args, **kwargs):
        super(Step, self).__init__(*args, **kwargs)
        self.convert = convert
        if sys is not None:
            self.sys = sys
        if meth is not None:
            self.meth = meth

    def __repr__(self):
        return '%s(%r, %r, %r)' % (
            self.__class__.__name__,
            self.ctx,
            self.sys,
            self.meth)

    def launch(self, *args, **kwargs):
        status = super(Step, self).launch(*args, **kwargs)
        if status != OK:
            raise get_exception(status, self.ctx.get_err())

    def step(self, *args, **kwargs):
        status = super(Step, self).step(*args, **kwargs)
        if status != OK:
            raise get_exception(status, self.ctx.get_err())

    def walk(self, *args, **kwargs):
        status = super(Step, self).walk(*args, **kwargs)
        if status != OK:
            raise get_exception(status, self.ctx.get_err())

class Operator(Operator):
    def __init__(self, ctx, sys=None, caps=None):
        super(Operator, self).__init__()
        self.convert = convert
        if sys is not None:
            self.sys = sys
        if caps is not None:
            self.caps = caps

    def __repr__(self):
        return '%s(%r, %r)' % (
            self.__class__.__name__,
            self.ctx,
            self.sys)

    def __call__(self, *args, **kwargs):
        (status, result) = super(Operator, self).__call__(*args, **kwargs)
        if status != OK:
            raise get_exception(status, self.ctx.get_err())
        return result

convert = TypeConvert()
convert.FullAddress = FullAddress
convert.Context = Context
convert.Method = Method
convert.CustomMethod = CustomMethod
convert.LinearMethod = LinearMethod
convert.PageTableMethod = PageTableMethod
convert.LookupMethod = LookupMethod
convert.MemoryArrayMethod = MemoryArrayMethod
convert.Range = Range
convert.Map = Map
convert.System = System
convert.Step = Step
convert.Operator = Operator

import inspect as _inspect
_values = globals().values()
if sys.version_info.major >= 3:
    _values = list(_values)
for _cls in _values:
    if not _inspect.isclass(_cls):
        continue
    for _name, _method in _inspect.getmembers(_cls,
                                              lambda x:
                                              (_inspect.ismethod(x),
                                               _inspect.isfunction(x))
                                              [sys.version_info.major >= 3]):
        if _method.__doc__:
            continue
        for _parent in _inspect.getmro(_cls)[1:]:
            if hasattr(_parent, _name):
                if _inspect.ismethod(_method):
                    _method.__func__.__doc__ = getattr(_parent, _name).__doc__
                else:
                    _method.__doc__ = getattr(_parent, _name).__doc__
                break

# Free up temporary variables
del _values, _cls, _name, _method, _parent, _inspect



#!/usr/bin/env python
# vim:sw=4 ts=4 et

from __future__ import print_function
import kdumpfile
import addrxlat
from sys import argv

tbl_names = ( 'PAGE', 'PTE', 'PMD', 'PUD', 'PGD' )

class kphysnote(object):
    def __init__(self, ctx, sys):
        self.ctx = ctx
        self.sys = sys

        map = sys.get_map(addrxlat.SYS_MAP_MACHPHYS_KPHYS)
        self.ident = True
        for range in map:
            if range.meth == addrxlat.SYS_METH_NONE:
                continue
            meth = sys.get_meth(range.meth)
            if meth.kind != addrxlat.LINEAR or meth.off != 0:
                self.ident = False
                break

    def note(self, fulladdr, fmt='{}'):
        if self.ident or fulladdr.addrspace == addrxlat.KPHYSADDR:
            return ''
        tmp = fulladdr.copy()
        try:
            tmp.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
            return fmt.format('{:x}'.format(tmp.addr))
        except (addrxlat.NotPresentError, addrxlat.NoDataError):
            return fmt.format('N/A')

def vtop(addr, ctx, sys):
    fulladdr = addrxlat.FullAddress(addrxlat.KVADDR, addr)
    print('{:16}  {:16}'.format('VIRTUAL', 'PHYSICAL'))
    try:
        fulladdr.conv(addrxlat.KPHYSADDR, ctx, sys)
        print('{:<16x}  {:<16x}\n'.format(addr, fulladdr.addr))
    except addrxlat.BaseException:
        print('{:<16x}  {:<16}\n'.format(addr, '---'))

    step = addrxlat.Step(ctx, sys)
    meth = sys.get_map(addrxlat.SYS_MAP_HW).search(addr)
    if meth == addrxlat.SYS_METH_NONE:
        meth = sys.get_map(addrxlat.SYS_MAP_KV_PHYS).search(addr)
    if meth == addrxlat.SYS_METH_NONE:
        print('NO METHOD')
        return

    step.meth = sys.get_meth(meth)
    step.launch(addr)

    note = kphysnote(ctx, sys)
    while step.remain:
        tbl = tbl_names[step.remain - 1]
        if step.remain > 1:
            addr = step.base.copy()
            addr.addr += step.idx[step.remain - 1] * step.elemsz
        else:
            addr = step.base
        remark = note.note(addr, ' ({})')
        print('{:>4}: {:16x}{}'.format(tbl, addr.addr, remark), end='')

        try:
            step.step()
            if step.remain and step.raw is not None:
                print(' => {:x}'.format(step.raw))
        except addrxlat.NotPresentError:
            print(' => {:x}  NOT PRESENT'.format(step.raw))
            return
    print()

if len(argv) != 3:
    print('Usage: {} <dumpfile> <virtual address>'.format(argv[0]))
    exit(1)

kdf = kdumpfile.kdumpfile(argv[1])
kdf.addrxlat_convert = addrxlat.convert
kdf.attr['addrxlat.ostype'] = 'linux'
sys = kdf.get_addrxlat_sys()
ctx = kdf.get_addrxlat_ctx()

addr = int(argv[2], base=0)
try:
    vtop(addr, ctx, sys)
except addrxlat.BaseException as e:
    print('Translation failed: {}'.format(e.message))



#!/usr/bin/env python
# vim:sw=4 ts=4 et

KDUMP_OK           = 0
KDUMP_ERR_SYSTEM   = 1
KDUMP_ERR_NOTIMPL  = 2
KDUMP_ERR_NODATA   = 3
KDUMP_ERR_CORRUPT  = 4
KDUMP_ERR_INVALID  = 5
KDUMP_ERR_EOF      = 6
KDUMP_ERR_NOKEY    = 7
KDUMP_ERR_BUSY     = 8
KDUMP_ERR_ADDRXLAT = 9

class KDumpBaseException(Exception):
    error = None

class OSErrorException(KDumpBaseException):
    error = KDUMP_ERR_SYSTEM

class NotImplementedException(KDumpBaseException):
    error = KDUMP_ERR_NOTIMPL

class NoDataException(KDumpBaseException):
    error = KDUMP_ERR_NODATA

class CorruptException(KDumpBaseException):
    error = KDUMP_ERR_CORRUPT

class InvalidException(KDumpBaseException):
    error = KDUMP_ERR_INVALID

class NoKeyException(KDumpBaseException):
    error = KDUMP_ERR_NOKEY

class EOFException(KDumpBaseException):
    error = KDUMP_ERR_EOF

class BusyException(KDumpBaseException):
    error = KDUMP_ERR_BUSY

class AddressTranslationException(KDumpBaseException):
    error = KDUMP_ERR_ADDRXLAT



from distutils.core import Command

class get_build_platlib(Command):
    """This command shows the platform build directory.
    """
    description = ("Show platform build directory")
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd = self.get_finalized_command('build')
        print(cmd.build_platlib)



#!/usr/bin/env python
# vim:sw=4 ts=4 et

# A version of examples/showxlat.c translated to Python.
# The script is not meant to be particularly useful, but it should
# demonstrate use of the libaddrxlat Python API.
#
# This file is in public domain.

from __future__ import print_function

import sys
import kdumpfile
import addrxlat

def addrspace_str(addrspace):
    if addrspace == addrxlat.KPHYSADDR:
        return 'KPHYSADDR'
    elif addrspace == addrxlat.MACHPHYSADDR:
        return 'MACHPHYSADDR'
    elif addrspace == addrxlat.KVADDR:
        return 'KVADDR'
    elif addrspace == addrxlat.NOADDR:
        return 'NOADDR'
    else:
        return '<addrspace {:d}>'.format(addrspace)

def fulladdr_str(addr):
    result = addrspace_str(addr.addrspace)
    if addr.addrspace != addrxlat.NOADDR:
        result += ':0x{:x}'.format(addr.addr)
    return result

def print_target_as(meth):
    print('  target_as={}'.format(addrspace_str(meth.target_as)))

def print_linear(meth):
    print('LINEAR')
    print_target_as(meth)
    print('  off=0x{:x}'.format(meth.off))

pte_formats = {
    addrxlat.PTE_NONE: 'none',
    addrxlat.PTE_PFN32: 'pfn32',
    addrxlat.PTE_PFN64: 'pfn64',
    addrxlat.PTE_AARCH64: 'aarch64',
    addrxlat.PTE_AARCH64_LPA: 'aarch64_lpa',
    addrxlat.PTE_AARCH64_LPA2: 'aarch64_lpa2',
    addrxlat.PTE_IA32: 'ia32',
    addrxlat.PTE_IA32_PAE: 'ia32_pae',
    addrxlat.PTE_X86_64: 'x86_64',
    addrxlat.PTE_S390X: 's390x',
    addrxlat.PTE_PPC64_LINUX_RPN30: 'ppc64_linux_rpn30',
}

def print_pgt(meth):
    print('PGT')
    print_target_as(meth)
    print('  root={}'.format(fulladdr_str(meth.root)))
    fmt = pte_formats.get(meth.pte_format, meth.pte_format)
    print('  pte_format={}'.format(fmt))
    print('  fields={}'.format(','.join(str(i) for i in meth.fields)))

def print_lookup(meth):
    print('LOOKUP')
    print_target_as(meth)
    print('  endoff=0x{:x}'.format(meth.endoff))
    for elem in meth.tbl:
        print('  {:x} -> {:x}'.format(elem[0], elem[1]))

def print_memarr(meth):
    print('MEMARR')
    print_target_as(meth)
    print('  base={}'.format(fulladdr_str(meth.base)))
    print('  shift={}'.format(meth.shift))
    print('  elemsz={}'.format(meth.elemsz))
    print('  valsz={}'.format(meth.valsz))

def print_meth(system, name):
    meth = system.get_meth(addrxlat.__dict__['SYS_METH_{}'.format(name)])
    if meth.kind == addrxlat.NOMETH:
        return

    print('METH_{}: '.format(name), end='')

    if meth.kind == addrxlat.CUSTOM:
        print('CUSTOM')
    elif meth.kind == addrxlat.LINEAR:
        print_linear(meth)
    elif meth.kind == addrxlat.PGT:
        print_pgt(meth)
    elif meth.kind == addrxlat.LOOKUP:
        print_lookup(meth)
    elif meth.kind == addrxlat.MEMARR:
        print_memarr(meth)
    else:
        print('<meth kind {:d}>'.format(meth.kind))

    print()

def print_map(system, name):
    print('MAP_{}:'.format(name))

    map = system.get_map(addrxlat.__dict__['SYS_MAP_{}'.format(name)])
    if map is None:
        return

    addr = 0
    for range in map:
        if range.meth in meth_names:
            name = meth_names[range.meth]
        else:
            name = '{:d}'.format(range.meth)
        print('{:x}-{:x}: {}'.format(addr, addr + range.endoff, name))
        addr += range.endoff + 1

def dump_addrxlat(ctx):
    system = k.get_addrxlat_sys()

    print_meth(system, 'PGT')
    print_meth(system, 'UPGT')
    print_meth(system, 'DIRECT')
    print_meth(system, 'KTEXT')
    print_meth(system, 'VMEMMAP')
    print_meth(system, 'RDIRECT')
    print_meth(system, 'MACHPHYS_KPHYS')
    print_meth(system, 'KPHYS_MACHPHYS')

    print_map(system, 'HW')
    print()
    print_map(system, 'KV_PHYS')
    print()
    print_map(system, 'KPHYS_DIRECT')
    print()
    print_map(system, 'MACHPHYS_KPHYS')
    print()
    print_map(system, 'KPHYS_MACHPHYS')

if __name__ == '__main__':
    if len(sys.argv) < 2 or len(sys.argv) > 3:
        print('Usage: {} <dumpfile> [<ostype>]'.format(sys.argv[0]),
              file=sys.stderr)
        exit(1)

    k = kdumpfile.kdumpfile(sys.argv[1])

    meth_names = {}
    for name, val in addrxlat.__dict__.items():
        if name.startswith('SYS_METH_') and name != 'SYS_METH_NUM':
            meth_names[val] = name[9:]

    if len(sys.argv) > 2:
        k.attr['addrxlat.ostype'] = sys.argv[2]

    dump_addrxlat(k)



#!/usr/bin/env python
# vim:sw=4 ts=4 et

import unittest
import addrxlat
import sys

if (sys.version_info.major >= 3):
    xrange = range

class TestAddressSpace(unittest.TestCase):
    names = (
        'KPHYSADDR',
        'MACHPHYSADDR',
        'KVADDR',
        'NOADDR',
    )

    def test_addrspace_name(self):
        for name in self.names:
            self.assertEqual(addrxlat.addrspace_name(addrxlat.__dict__[name]),
                             name)

class TestFullAddress(unittest.TestCase):
    def test_fulladdr_defaults(self):
        addr = addrxlat.FullAddress()
        self.assertEqual(addr.addrspace, addrxlat.NOADDR)
        self.assertEqual(addr.addr, 0)

    def test_fulladdr_addrspace(self):
        addr = addrxlat.FullAddress(addrspace=addrxlat.KVADDR)
        self.assertEqual(addr.addrspace, addrxlat.KVADDR)
        self.assertEqual(addr.addr, 0)

    def test_fulladdr_addr(self):
        addr = addrxlat.FullAddress(addr=0xabcd)
        self.assertEqual(addr.addrspace, addrxlat.NOADDR)
        self.assertEqual(addr.addr, 0xabcd)

    def test_fulladdr_init_pos(self):
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0xabcd)
        self.assertEqual(addr.addrspace, addrxlat.KVADDR)
        self.assertEqual(addr.addr, 0xabcd)

    def test_fulladdr_init_kwarg(self):
        addr = addrxlat.FullAddress(addr=0xabcd, addrspace=addrxlat.KVADDR)
        self.assertEqual(addr.addrspace, addrxlat.KVADDR)
        self.assertEqual(addr.addr, 0xabcd)

    def test_fulladdr_eq(self):
        addr1 = addrxlat.FullAddress(addrxlat.KVADDR, 0x1234)
        addr2 = addrxlat.FullAddress(addrxlat.KVADDR, 0x1234)
        self.assertEqual(addr1, addr2)

    def test_fulladdr_noteq(self):
        addr1 = addrxlat.FullAddress(addrxlat.KVADDR, 0x1234)
        addr2 = addrxlat.FullAddress(addrxlat.KVADDR, 0xabcd)
        self.assertNotEqual(addr1, addr2)

    def test_fulladdr_copy(self):
        addr1 = addrxlat.FullAddress(addrxlat.KVADDR, 0x1234)
        addr2 = addr1.copy()
        self.assertIsInstance(addr2, addrxlat.FullAddress)
        self.assertEqual(addr1.addrspace, addr2.addrspace)
        self.assertEqual(addr1.addr, addr2.addr)

    def test_fulladdr_copy_subclass(self):
        class myfulladdr(addrxlat.FullAddress):
            pass

        addr1 = myfulladdr(addrxlat.KVADDR, 0x1234)
        addr2 = addr1.copy()
        self.assertIsInstance(addr2, myfulladdr)

class TestContext(unittest.TestCase):
    def test_err(self):
        ctx = addrxlat.Context()
        ctx.clear_err()
        self.assertIs(ctx.get_err(), None)
        status = ctx.err(addrxlat.ERR_CUSTOM_BASE, 'An error message')
        self.assertEqual(status, addrxlat.ERR_CUSTOM_BASE)
        self.assertEqual(ctx.get_err(), 'An error message')
        ctx.clear_err()
        self.assertIs(ctx.get_err(), None)

class TestMethod(unittest.TestCase):
    def test_meth_defaults(self):
        meth = addrxlat.Method(addrxlat.NOMETH)
        self.assertEqual(meth.kind, addrxlat.NOMETH)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        for i in xrange(len(meth.param)):
            self.assertEqual(meth.param[i], 0)

    def test_meth_readonly_kind(self):
        meth = addrxlat.Method(addrxlat.NOMETH)
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.LINEAR

    def test_meth_target_as(self):
        meth = addrxlat.Method(addrxlat.NOMETH, addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.NOMETH)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        for i in xrange(len(meth.param)):
            self.assertEqual(meth.param[i], 0)

    def test_meth_param(self):
        meth = addrxlat.Method(addrxlat.NOMETH, param=(0, 1, 2, 3))
        self.assertGreaterEqual(len(meth.param), 4)
        self.assertEqual(meth.kind, addrxlat.NOMETH)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        for i in xrange(4):
            self.assertEqual(meth.param[i], i)
        for i in xrange(4, len(meth.param)):
            self.assertEqual(meth.param[i], 0)

        for i in xrange(4):
            meth.param[i] += 1
        self.assertEqual(meth.kind, addrxlat.NOMETH)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        for i in xrange(4):
            self.assertEqual(meth.param[i], i + 1)
        for i in xrange(4, len(meth.param)):
            self.assertEqual(meth.param[i], 0)

        with self.assertRaises(OverflowError):
            meth.param[0] = 999
        with self.assertRaises(OverflowError):
            meth.param[0] = -1

    def test_custom_defaults(self):
        meth = addrxlat.CustomMethod()
        self.assertEqual(meth.kind, addrxlat.CUSTOM)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)

    def test_custom_readonly_kind(self):
        meth = addrxlat.CustomMethod()
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.NOMETH

    def test_custom_target_as(self):
        meth = addrxlat.CustomMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.CUSTOM)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)

    def test_custom_notimpl(self):
        meth = addrxlat.CustomMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.CUSTOM)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        ctx = addrxlat.Context()
        step = addrxlat.Step(ctx=ctx, meth=meth)
        with self.assertRaisesRegex(BaseException, "NULL callback"):
            meth.cb_first_step(step, 0x1234)
        with self.assertRaisesRegex(BaseException, "NULL callback"):
            meth.cb_next_step(step)

    def test_linear_defaults(self):
        meth = addrxlat.LinearMethod()
        self.assertEqual(meth.kind, addrxlat.LINEAR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.off, 0)

    def test_linear_readonly_kind(self):
        meth = addrxlat.LinearMethod()
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.NOMETH

    def test_linear_target_as(self):
        meth = addrxlat.LinearMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.LINEAR)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.off, 0)

    def test_linear_off(self):
        meth = addrxlat.LinearMethod(off=addrxlat.ADDR_MAX - 0x1234)
        self.assertEqual(meth.kind, addrxlat.LINEAR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.off, addrxlat.ADDR_MAX - 0x1234)

    def test_linear_neg_off(self):
        meth = addrxlat.LinearMethod(off=-addrxlat.ADDR_MAX)
        self.assertEqual(meth.kind, addrxlat.LINEAR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.off, 1)

    def test_linear_param(self):
        meth = addrxlat.LinearMethod(off=0x1234)
        param = tuple(meth.param)
        self.assertLess(param.index(0x34), len(meth.param))
        meth.param = (0xff,) * len(meth.param)
        self.assertNotEqual(meth.off, 0x1234)

    def test_pgt_defaults(self):
        meth = addrxlat.PageTableMethod()
        self.assertEqual(meth.kind, addrxlat.PGT)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertIs(meth.root, None)
        self.assertEqual(meth.pte_format, addrxlat.PTE_NONE)
        self.assertEqual(meth.fields, tuple())

    def test_pgt_readonly_kind(self):
        meth = addrxlat.PageTableMethod()
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.NOMETH

    def test_pgt_target_as(self):
        meth = addrxlat.PageTableMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.PGT)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        self.assertIs(meth.root, None)
        self.assertEqual(meth.pte_format, addrxlat.PTE_NONE)
        self.assertEqual(meth.fields, tuple())

    def test_pgt_root(self):
        root = addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0x1000)
        meth = addrxlat.PageTableMethod(root=root)
        self.assertEqual(meth.kind, addrxlat.PGT)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.root, root)
        self.assertEqual(meth.pte_format, addrxlat.PTE_NONE)
        self.assertEqual(meth.fields, tuple())

    def test_pgt_pte_format(self):
        meth = addrxlat.PageTableMethod(pte_format=addrxlat.PTE_PFN32)
        self.assertEqual(meth.kind, addrxlat.PGT)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertIs(meth.root, None)
        self.assertEqual(meth.pte_format, addrxlat.PTE_PFN32)
        self.assertEqual(meth.fields, tuple())

    def test_pgt_fields(self):
        meth = addrxlat.PageTableMethod(fields=(1, 2, 3))
        self.assertEqual(meth.kind, addrxlat.PGT)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertIs(meth.root, None)
        self.assertEqual(meth.pte_format, addrxlat.PTE_NONE)
        self.assertEqual(meth.fields, (1, 2, 3))

        meth.fields = (4, 5, 6)
        self.assertEqual(meth.fields, (4, 5, 6))
        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            meth.fields = None
        with self.assertRaisesRegex(ValueError,
                                     'more than [0-9]+ address fields'):
            meth.fields = (0,) * (addrxlat.FIELDS_MAX + 1)

    def test_lookup_defaults(self):
        meth = addrxlat.LookupMethod()
        self.assertEqual(meth.kind, addrxlat.LOOKUP)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.endoff, 0)
        self.assertEqual(meth.tbl, tuple())

    def test_lookup_readonly_kind(self):
        meth = addrxlat.LookupMethod()
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.NOMETH

    def test_lookup_target_as(self):
        meth = addrxlat.LookupMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.LOOKUP)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.endoff, 0)
        self.assertEqual(meth.tbl, tuple())

    def test_lookup_endoff(self):
        meth = addrxlat.LookupMethod(endoff=0x1234)
        self.assertEqual(meth.kind, addrxlat.LOOKUP)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.endoff, 0x1234)
        self.assertEqual(meth.tbl, tuple())

    def test_lookup_tbl(self):
        meth = addrxlat.LookupMethod(tbl=((0, 100), (200, 300)))
        self.assertEqual(meth.kind, addrxlat.LOOKUP)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.endoff, 0)
        self.assertEqual(meth.tbl, ((0, 100), (200, 300)))

        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            meth.tbl = None
        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            meth.tbl = (None,)
        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            meth.tbl = 1
        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            meth.tbl = (1,)
        with self.assertRaisesRegex(ValueError, 'must be integer pairs'):
            meth.tbl = ((),)
        with self.assertRaisesRegex(ValueError, 'must be integer pairs'):
            meth.tbl = ((1,),)
        with self.assertRaisesRegex(ValueError, 'must be integer pairs'):
            meth.tbl = ((1, 2, 3),)
        with self.assertRaisesRegex(TypeError, 'must be.* a .*number'):
            meth.tbl = ((None, None),)
        with self.assertRaisesRegex(TypeError, 'must be.* a .*number'):
            meth.tbl = ((1, None),)
        with self.assertRaisesRegex(TypeError, 'must be.* a .*number'):
            meth.tbl = ((None, 1),)

    def test_memarr_defaults(self):
        meth = addrxlat.MemoryArrayMethod()
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.base, None)
        self.assertEqual(meth.shift, 0)
        self.assertEqual(meth.elemsz, 0)
        self.assertEqual(meth.valsz, 0)

    def test_memarr_readonly_kind(self):
        meth = addrxlat.MemoryArrayMethod()
        with self.assertRaises(AttributeError):
            meth.kind = addrxlat.NOMETH

    def test_memarr_target_as(self):
        meth = addrxlat.MemoryArrayMethod(addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.MACHPHYSADDR)
        self.assertEqual(meth.base, None)
        self.assertEqual(meth.shift, 0)
        self.assertEqual(meth.elemsz, 0)
        self.assertEqual(meth.valsz, 0)

    def test_memarr_base(self):
        base = addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0x1234)
        meth = addrxlat.MemoryArrayMethod(base=base)
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.base, base)
        self.assertEqual(meth.shift, 0)
        self.assertEqual(meth.elemsz, 0)
        self.assertEqual(meth.valsz, 0)

    def test_memarr_shift(self):
        meth = addrxlat.MemoryArrayMethod(shift=3)
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.base, None)
        self.assertEqual(meth.shift, 3)
        self.assertEqual(meth.elemsz, 0)
        self.assertEqual(meth.valsz, 0)

    def test_memarr_elemsz(self):
        meth = addrxlat.MemoryArrayMethod(elemsz=12)
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.base, None)
        self.assertEqual(meth.shift, 0)
        self.assertEqual(meth.elemsz, 12)
        self.assertEqual(meth.valsz, 0)

    def test_memarr_valsz(self):
        meth = addrxlat.MemoryArrayMethod(valsz=8)
        self.assertEqual(meth.kind, addrxlat.MEMARR)
        self.assertEqual(meth.target_as, addrxlat.NOADDR)
        self.assertEqual(meth.base, None)
        self.assertEqual(meth.shift, 0)
        self.assertEqual(meth.elemsz, 0)
        self.assertEqual(meth.valsz, 8)

class TestRange(unittest.TestCase):
    def test_range_defaults(self):
        range = addrxlat.Range()
        self.assertEqual(range.endoff, 0)
        self.assertEqual(range.meth, addrxlat.SYS_METH_NONE)

    def test_range_endoff(self):
        range = addrxlat.Range(endoff=0x1234)
        self.assertEqual(range.endoff, 0x1234)
        self.assertEqual(range.meth, addrxlat.SYS_METH_NONE)

    def test_range_meth(self):
        meth = addrxlat.SYS_METH_PGT
        range = addrxlat.Range(meth=meth)
        self.assertEqual(range.endoff, 0)
        self.assertIs(range.meth, addrxlat.SYS_METH_PGT)

class TestMap(unittest.TestCase):
    def test_map_defaults(self):
        map = addrxlat.Map()
        self.assertEqual(len(map), 0)

    def test_map_set(self):
        map = addrxlat.Map()
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_PGT))
        self.assertEqual(len(map), 2)
        self.assertEqual(map[0].endoff, 0xffff)
        self.assertEqual(map[0].meth, addrxlat.SYS_METH_PGT)
        self.assertEqual(map[1].meth, addrxlat.SYS_METH_NONE)

    def test_map_search(self):
        map = addrxlat.Map()
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_PGT))
        meth2 = map.search(0)
        self.assertEqual(meth2, addrxlat.SYS_METH_PGT)
        meth2 = map.search(0x10000)
        self.assertIs(meth2, addrxlat.SYS_METH_NONE)

    def test_map_copy(self):
        map = addrxlat.Map()
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_PGT))
        self.assertEqual(len(map), 2)
        map2 = map.copy()
        self.assertNotEqual(map2, map)
        self.assertEqual(len(map2), 2)
        self.assertEqual(map2[0].endoff, 0xffff)
        self.assertEqual(map2[0].meth, addrxlat.SYS_METH_PGT)
        self.assertEqual(map2[1].endoff, map[1].endoff)
        self.assertIs(map2[1].meth, addrxlat.SYS_METH_NONE)

class TestSystem(unittest.TestCase):
    def test_sys_defaults(self):
        sys = addrxlat.System()
        for i in xrange(addrxlat.SYS_MAP_NUM):
            map = sys.get_map(i)
            self.assertIs(map, None)
        for i in xrange(addrxlat.SYS_METH_NUM):
            meth = sys.get_meth(i)
            self.assertEqual(meth.kind, addrxlat.NOMETH)

    def test_sys_map(self):
        sys = addrxlat.System()
        newmap = addrxlat.Map()
        for mapidx in xrange(addrxlat.SYS_MAP_NUM):
            sys.set_map(mapidx, newmap)
            for i in xrange(mapidx + 1):
                map = sys.get_map(i)
                self.assertEqual(map, newmap)
            for i in xrange(mapidx + 1, addrxlat.SYS_MAP_NUM):
                map = sys.get_map(i)
                self.assertIs(map, None)
        for i in xrange(addrxlat.SYS_METH_NUM):
            meth = sys.get_meth(i)
            self.assertEqual(meth.kind, addrxlat.NOMETH)

    def test_sys_meth(self):
        sys = addrxlat.System()
        newdesc = addrxlat.LinearMethod(0)
        for i in xrange(addrxlat.SYS_MAP_NUM):
            map = sys.get_map(i)
            self.assertIs(map, None)
        for methidx in xrange(addrxlat.SYS_METH_NUM):
            sys.set_meth(methidx, newdesc)
            for i in xrange(methidx):
                meth = sys.get_meth(i)
                self.assertEqual(meth, newdesc)
            for i in xrange(methidx + 1, addrxlat.SYS_METH_NUM):
                meth = sys.get_meth(i)
                self.assertEqual(meth.kind, addrxlat.NOMETH)

class TestStep(unittest.TestCase):
    def setUp(self):
        self.ctx = addrxlat.Context()

    def test_step_defaults(self):
        step = addrxlat.Step(self.ctx)
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_sys(self):
        sys = addrxlat.System()
        step = addrxlat.Step(self.ctx, sys=sys)
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, sys)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_meth(self):
        meth = addrxlat.LinearMethod()
        step = addrxlat.Step(self.ctx, meth=meth)
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, meth)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_remain(self):
        step = addrxlat.Step(self.ctx)
        step.remain = 3
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 3)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_elemsz(self):
        step = addrxlat.Step(self.ctx)
        step.elemsz = 8
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 8)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_base(self):
        step = addrxlat.Step(self.ctx)
        base = addrxlat.FullAddress(addrxlat.KVADDR, 0xabcd)
        step.base = base
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, base)
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_base_addrspace(self):
        step = addrxlat.Step(self.ctx)
        step.base = addrxlat.FullAddress(addrxlat.NOADDR, 0)
        step.base.addrspace = addrxlat.KVADDR
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertEqual(step.base, addrxlat.FullAddress(addrxlat.KVADDR, 0))
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_base_addr(self):
        step = addrxlat.Step(self.ctx)
        step.base = addrxlat.FullAddress(addrxlat.NOADDR, 0)
        step.base.addr = 0x1234
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertEqual(step.base, addrxlat.FullAddress(addrxlat.NOADDR, 0x1234))
        self.assertIs(step.raw, None)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_raw(self):
        step = addrxlat.Step(self.ctx)
        with self.assertRaisesRegex(TypeError, 'cannot be changed'):
            step.raw = 0xabcd
        meth = addrxlat.PageTableMethod()
        step.meth = meth
        step.raw = 0xabcd
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, meth)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertEqual(step.raw, 0xabcd)
        idx = (0,) * (addrxlat.FIELDS_MAX + 1)
        self.assertEqual(step.idx, idx)

    def test_step_idx(self):
        step = addrxlat.Step(self.ctx)
        idx = (1, 2, 3, 4)
        step.idx = idx
        self.assertIs(step.ctx, self.ctx)
        self.assertIs(step.sys, None)
        self.assertIs(step.meth, None)
        self.assertEqual(step.remain, 0)
        self.assertEqual(step.elemsz, 0)
        self.assertIs(step.base, None)
        self.assertIs(step.raw, None)
        idx = idx + (0,) * (addrxlat.FIELDS_MAX + 1 - len(idx))
        self.assertEqual(step.idx, idx)

        with self.assertRaisesRegex(TypeError, 'not a sequence'):
            step.idx = None
        with self.assertRaisesRegex(ValueError, 'more than [0-9]+ indices'):
            step.idx = (0,) * (addrxlat.FIELDS_MAX + 2)
        with self.assertRaisesRegex(TypeError, 'must be.* a .*number'):
            step.idx = (None,)

class TestOperator(unittest.TestCase):
    def setUp(self):
        self.ctx = addrxlat.Context()

    def test_op_defaults(self):
        op = addrxlat.Operator(self.ctx)
        self.assertIs(op.ctx, self.ctx)
        self.assertIs(op.sys, None)
        self.assertEqual(op.caps, 0)

    def test_op_sys(self):
        sys = addrxlat.System()
        op = addrxlat.Operator(self.ctx, sys=sys)
        self.assertIs(op.ctx, self.ctx)
        self.assertEqual(op.sys, sys)
        self.assertEqual(op.caps, 0)

    def test_op_caps(self):
        op = addrxlat.Operator(self.ctx, caps=addrxlat.CAPS(addrxlat.KVADDR))
        self.assertIs(op.ctx, self.ctx)
        self.assertIs(op.sys, None)
        self.assertEqual(op.caps, addrxlat.CAPS(addrxlat.KVADDR))

#
# Test translation system has the following layout:
#
# MAP_HW:
# 0-ffff: PGT root=MACHPHYSADDR:0
# 10000-ffffffffffffffff: NONE
#
# MAP_KV_PHYS:
# 0-1fff: DIRECT off=0x1000
# 2000-3fff: LOOKUP
# 4000-5fff: MEMARR base=KVADDR:0
# 6000-ffff: PGT
# 10000-ffffffffffffffff: NONE
#
# MAP_KPHYS_DIRECT:
# 0-fff: NONE
# 1000-2fff: RDIRECT
# 3000-ffffffffffffffff: NONE
#
# MAP_MACHPHYS_KPHYS:
# 0-ffff: NONE
# 10000-1ffff: KPHYS_MACHPHYS
# 20000-ffffffffffffffff: NONE
#
# MAP_KPHYS_MACHPHYS:
# 0-ffff: MACHPHYS_KPHYS
# 10000-ffffffffffffffff: NONE
#
class TestTranslation(unittest.TestCase):
    def setUp(self):
        def get_page(addr):
            # Page table level 2 @ 0
            if addr.addr == 0x10000:
                return (bytearray((0x00, 0x00, 0x01, 0x01)),
                        addrxlat.BIG_ENDIAN)
            # Page table level 1 @ 0x65
            if addr.addr == 0x10100 + 0x65 * 4:
                return (bytearray((0x00, 0x00, 0x01, 0xc0)),
                        addrxlat.BIG_ENDIAN)
            # Page table level 1 @ 0x41
            if addr.addr == 0x10100 + 0x41 * 4:
                return (bytearray((0x00, 0x00, 0x01, 0xa9)),
                        addrxlat.BIG_ENDIAN)
            # Memory array at 0x40
            if addr.addr == 0x11000 + 0x40 * 4:
                return (bytearray((0x00, 0x00, 0x00, 0xaa)),
                        addrxlat.BIG_ENDIAN)

        def read_caps():
            return addrxlat.CAPS(addrxlat.MACHPHYSADDR)

        self.ctx = addrxlat.Context()
        self.ctx.cb_read_caps = read_caps
        self.ctx.cb_get_page = get_page
        self.sys = addrxlat.System()

        map = addrxlat.Map()
        self.sys.set_map(addrxlat.SYS_MAP_HW, map)
        meth = addrxlat.PageTableMethod(addrxlat.MACHPHYSADDR)
        meth.root = addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0x10000)
        meth.pte_format = addrxlat.PTE_PFN32
        meth.fields = (8, 8, 8)
        self.sys.set_meth(addrxlat.SYS_METH_PGT, meth)
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_PGT))

        map = addrxlat.Map()
        self.sys.set_map(addrxlat.SYS_MAP_KV_PHYS, map)
        meth = addrxlat.LinearMethod(addrxlat.KPHYSADDR, 0x1000)
        self.sys.set_meth(addrxlat.SYS_METH_DIRECT, meth)
        map.set(0, addrxlat.Range(0x1fff, addrxlat.SYS_METH_DIRECT))
        meth = addrxlat.LookupMethod(addrxlat.KPHYSADDR)
        meth.endoff = 0xff
        meth.tbl = ((0x2000, 0xfa00), (0x3000, 0xfb00), (0x3100, 0xff00))
        self.sys.set_meth(addrxlat.SYS_METH_CUSTOM, meth)
        map.set(0x2000, addrxlat.Range(0x1fff, addrxlat.SYS_METH_CUSTOM))
        meth = addrxlat.MemoryArrayMethod(addrxlat.KPHYSADDR)
        meth.base = addrxlat.FullAddress(addrxlat.KVADDR, 0)
        meth.shift = 8
        meth.elemsz = 4
        meth.valsz = 4
        self.sys.set_meth(addrxlat.SYS_METH_CUSTOM + 1, meth)
        map.set(0x4000, addrxlat.Range(0x1fff, addrxlat.SYS_METH_CUSTOM + 1))
        map.set(0x6000, addrxlat.Range(0x9fff, addrxlat.SYS_METH_PGT))

        map = addrxlat.Map()
        self.sys.set_map(addrxlat.SYS_MAP_KPHYS_DIRECT, map)
        meth = addrxlat.LinearMethod(addrxlat.KVADDR, -0x1000)
        self.sys.set_meth(addrxlat.SYS_METH_RDIRECT, meth)
        map.set(0x1000, addrxlat.Range(0x1fff, addrxlat.SYS_METH_RDIRECT))

        map = addrxlat.Map()
        self.sys.set_map(addrxlat.SYS_MAP_MACHPHYS_KPHYS, map)
        meth = addrxlat.LinearMethod(addrxlat.KPHYSADDR, -0x10000)
        self.sys.set_meth(addrxlat.SYS_METH_MACHPHYS_KPHYS, meth)
        map.set(0x10000, addrxlat.Range(0xffff, addrxlat.SYS_METH_MACHPHYS_KPHYS))

        map = addrxlat.Map()
        self.sys.set_map(addrxlat.SYS_MAP_KPHYS_MACHPHYS, map)
        meth = addrxlat.LinearMethod(addrxlat.MACHPHYSADDR, 0x10000)
        self.sys.set_meth(addrxlat.SYS_METH_KPHYS_MACHPHYS, meth)
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_KPHYS_MACHPHYS))

    def test_fulladdr_conv_kphys_machphys(self):
        "KPHYS -> MACHPHYS using offset"
        addr = addrxlat.FullAddress(addrxlat.KPHYSADDR, 0x2345)
        addr.conv(addrxlat.MACHPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0x12345))

    def test_fulladdr_fail_kphys_machphys(self):
        "KPHYS -> MACHPHYS out of bounds"
        addr = addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xf4255)
        with self.assertRaisesRegex(addrxlat.NoMethodError, 'No way to translate'):
            addr.conv(addrxlat.MACHPHYSADDR, self.ctx, self.sys)

    def test_fulladdr_conv_machphys_kphys(self):
        "MACHPHYS -> KPHYS using offset"
        addr = addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0x1abcd)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xabcd))

    def test_fulladdr_fail_machphys_kphys(self):
        "MACHPHYS -> KPHYS out of bounds"
        addr = addrxlat.FullAddress(addrxlat.MACHPHYSADDR, 0xabcd)
        with self.assertRaisesRegex(addrxlat.NoMethodError, 'No way to translate'):
            addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)

    def test_fulladdr_conv_direct(self):
        "KV -> KPHYS using directmap"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x1234)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0x2234))

    def test_fulladdr_conv_lookup(self):
        "KV -> KPHYS using lookup"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x2055)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xfa55))

    def test_fulladdr_conv_memarr(self):
        "KV -> KPHYS using memory array"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x4055)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xaa55))

    def test_fulladdr_conv_memarr_pgt(self):
        "KV -> KPHYS using fallback from memory array to page tables"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x4155)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xa955))

    def test_fulladdr_fail_memarr(self):
        "KV -> KPHYS using memory array returns None"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x4255)
        with self.assertRaisesRegex(addrxlat.NoMethodError, 'Callback returned None'):
            addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)

    def test_fulladdr_conv_pgt(self):
        "KV -> KPHYS using page tables"
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x6502)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0xc002))

    def test_fulladdr_conv_rdirect(self):
        "KPHYS -> KV using reverse directmap"
        addr = addrxlat.FullAddress(addrxlat.KPHYSADDR, 0x2345)
        addr.conv(addrxlat.KVADDR, self.ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KVADDR, 0x1345))

    def test_op_direct(self):
        "Operator using directmap"
        class hexop(addrxlat.Operator):
            def __init__(self, prefix='', *args, **kwargs):
                super(hexop, self).__init__(*args, **kwargs)
                self.prefix = prefix

            def callback(self, addr):
                return '{}{:x}'.format(self.prefix, addr.addr)

        myop = hexop(ctx=self.ctx, sys=self.sys, caps=addrxlat.CAPS(addrxlat.KPHYSADDR), prefix='0x')
        result = myop(addrxlat.FullAddress(addrxlat.KVADDR, 0xabc))
        self.assertEqual(result, '0x1abc')

    def test_subclass_memarr(self):
        "KV -> KPHYS using memory array and a subclass"

        class mycontext(addrxlat.Context):
            def __init__(self, *args, **kwargs):
                super(mycontext, self).__init__(*args, **kwargs)
            def cb_read_caps(self):
                return addrxlat.CAPS(addrxlat.MACHPHYSADDR)
            def cb_get_page(self, addr):
                # Memory array at 0x40
                if addr.addr == 0x11000 + 0x40 * 4:
                    return (bytearray((0x00, 0x00, 0x00, 0x12)),
                            addrxlat.BIG_ENDIAN)

        ctx = mycontext()
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x4034)
        addr.conv(addrxlat.KPHYSADDR, ctx, self.sys)
        self.assertEqual(addr, addrxlat.FullAddress(addrxlat.KPHYSADDR, 0x1234))

class TestCustom(unittest.TestCase):
    def setUp(self):
        self.ctx = addrxlat.Context()

        def first_step(step, addr):
            step.base = addrxlat.FullAddress(addrxlat.NOADDR, 0xabcdef)
            step.idx = (addr & 0xff, addr >> 8)
            step.remain = 2

        def next_step(step):
            step.base.addr = 0x123456 + step.idx[1]
            step.elemsz = 0x100

        self.meth = addrxlat.CustomMethod()
        self.meth.target_as = addrxlat.KPHYSADDR
        self.meth.cb_first_step = first_step
        self.meth.cb_next_step = next_step

        import _test_addrxlat
        self.meth_ext = _test_addrxlat.getCustomMethod(addrxlat.convert)
        self.assertEqual(self.meth_ext.kind, addrxlat.CUSTOM)
        self.assertEqual(self.meth_ext.target_as, addrxlat.NOADDR)
        self.meth_ext.target_as = addrxlat.KPHYSADDR

        self.meth_extmod = _test_addrxlat.getCustomMethod(addrxlat.convert)
        self.meth_extmod.target_as = addrxlat.KPHYSADDR
        self.meth_extmod.cb_next_step = next_step

    def test_customdesc_cb(self):
        step = addrxlat.Step(ctx=self.ctx, meth=self.meth)
        self.assertIs(step.base, None)
        self.meth.cb_first_step(step, 0x1234)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0xabcdef)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)
        self.meth.cb_next_step(step)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0x123456 + 0x12)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)

    def test_customdesc_conv(self):
        sys = addrxlat.System()
        map = addrxlat.Map()
        sys.set_meth(addrxlat.SYS_METH_CUSTOM, self.meth)
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_CUSTOM))
        sys.set_map(addrxlat.SYS_MAP_KV_PHYS, map)
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x2345)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, sys)
        self.assertEqual(addr.addrspace, addrxlat.KPHYSADDR)
        self.assertEqual(addr.addr, 0x123456 + 0x4523)

    def test_customdesc_ext_cb(self):
        step = addrxlat.Step(ctx=self.ctx, meth=self.meth_ext)
        self.assertIs(step.base, None)
        self.meth_ext.cb_first_step(step, 0x1234)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0x4d795f4d61676963)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)
        self.meth_ext.cb_next_step(step)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0x4d61676963546f6f + 0x12)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)

    def test_customdesc_ext_conv(self):
        sys = addrxlat.System()
        map = addrxlat.Map()
        sys.set_meth(addrxlat.SYS_METH_CUSTOM, self.meth_ext)
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_CUSTOM))
        sys.set_map(addrxlat.SYS_MAP_KV_PHYS, map)
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x2345)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, sys)
        self.assertEqual(addr.addrspace, addrxlat.KPHYSADDR)
        self.assertEqual(addr.addr, 0x4d61676963546f6f + 0x4523)

    def test_customdesc_extmod_cb(self):
        step = addrxlat.Step(ctx=self.ctx, meth=self.meth_extmod)
        self.assertIs(step.base, None)
        self.meth_extmod.cb_first_step(step, 0x1234)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0x4d795f4d61676963)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)
        self.meth_extmod.cb_next_step(step)
        self.assertEqual(step.base.addrspace, addrxlat.NOADDR)
        self.assertEqual(step.base.addr, 0x123456 + 0x12)
        self.assertEqual(step.idx[0], 0x34)
        self.assertEqual(step.idx[1], 0x12)

    def test_customdesc_extmod_conv(self):
        sys = addrxlat.System()
        map = addrxlat.Map()
        sys.set_meth(addrxlat.SYS_METH_CUSTOM, self.meth_extmod)
        map.set(0, addrxlat.Range(0xffff, addrxlat.SYS_METH_CUSTOM))
        sys.set_map(addrxlat.SYS_MAP_KV_PHYS, map)
        addr = addrxlat.FullAddress(addrxlat.KVADDR, 0x2345)
        addr.conv(addrxlat.KPHYSADDR, self.ctx, sys)
        self.assertEqual(addr.addrspace, addrxlat.KPHYSADDR)
        self.assertEqual(addr.addr, 0x123456 + 0x4523)

if __name__ == '__main__':
    unittest.main()



from distutils.core import setup, Extension
import libtoolize
import disthelpers
import os

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser

cfg = ConfigParser()
cfg.read('setup.cfg')
srcdir = cfg.get('kdumpfile', 'srcdir')
top_builddir = cfg.get('kdumpfile', 'top_builddir')
include_dir=os.path.join(top_builddir, 'include')
addrxlat_la = os.path.join(
    top_builddir, 'src', 'addrxlat', 'libaddrxlat.la')
kdumpfile_la = os.path.join(
    top_builddir, 'src', 'kdumpfile', 'libkdumpfile.la')

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+) ",
    "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)  ",
    "Programming Language :: Python",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Software Development :: Debuggers "]

setup(name='libkdumpfile',
      version=cfg.get('kdumpfile', 'version'),
      description='Python bindings for libkdumpfile',
      author='Petr Tesarik',
      author_email='petr@tesarici.cz',
      url='https://github.com/ptesarik/libkdumpfile',
      packages=['addrxlat', 'kdumpfile'],
      package_dir={'': srcdir},
      ext_modules=[
          Extension('_addrxlat', [os.path.join(srcdir, 'addrxlat.c')],
                    include_dirs=[include_dir],
                    extra_objects=[addrxlat_la]),
          Extension('_kdumpfile', [os.path.join(srcdir, 'kdumpfile.c')],
                    include_dirs=[include_dir],
                    extra_objects=[kdumpfile_la]),
      ],
      cmdclass={
          'build_ext': libtoolize.build_ext,
          'install_lib': libtoolize.install_lib,
          'get_build_platlib': disthelpers.get_build_platlib,
      },
)



#!/usr/bin/env python

from _kdumpfile import *
from addrxlat import convert as _convert
import sys

class kdumpfile(kdumpfile):
    def __init__(self, *args, **kwargs):
        self.addrxlat_convert = _convert

import inspect as _inspect
_values = globals().values()
if sys.version_info.major >= 3:
    _values = list(_values)
for _cls in _values:
    if not _inspect.isclass(_cls):
        continue
    for _name, _method in _inspect.getmembers(_cls,
                                              lambda x:
                                              (_inspect.ismethod(x),
                                               _inspect.isfunction(x))
                                              [sys.version_info.major >= 3]):
        if _method.__doc__:
            continue
        for _parent in _inspect.getmro(_cls)[1:]:
            if hasattr(_parent, _name):
                if _inspect.ismethod(_method):
                    _method.__func__.__doc__ = getattr(_parent, _name).__doc__
                else:
                    _method.__doc__ = getattr(_parent, _name).__doc__
                break

# Free up temporary variables
del _values, _cls, _name, _method, _parent, _inspect



"""libtool

Build and install extensions using libtool.
"""

import os
from distutils import log
from distutils.util import split_quoted
from distutils.dir_util import mkpath
from distutils.file_util import copy_file
from distutils.spawn import spawn
from distutils.errors import DistutilsFileError

from distutils.command.build_ext import build_ext as _build_ext
class build_ext(_build_ext):
    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.libtool = None
        self.pyexecdir = None

    def finalize_options(self):
        _build_ext.finalize_options(self)

        if self.libtool is None:
            self.libtool = 'libtool'
        if self.pyexecdir is None:
            import sysconfig
            self.pyexecdir = sysconfig.get_path('platlib')
        self.cmd_libtool = split_quoted(self.libtool)

    def libtoolize(self, key, mode, *args):
        val = getattr(self.compiler, key)
        val = val[:1] + list(args) + val[1:]
        setattr(self.compiler, key,
                self.cmd_libtool + [ '--mode='+mode ] + val)

    def build_extensions(self):
        self.compiler.obj_extension = '.lo'
        self.libtoolize('compiler', 'compile')
        self.libtoolize('compiler_so', 'compile')
        self.libtoolize('linker_exe', 'link')
        self.libtoolize('linker_so', 'link',
                        '-module', '-avoid-version',
                        '-export-symbols-regex', 'init.*|PyInit_.*',
                        '-rpath', self.pyexecdir)
        _build_ext.build_extensions(self)

from distutils.command.install_lib import install_lib as _install_lib
class install_lib(_install_lib):
    def initialize_options(self):
        _install_lib.initialize_options(self)
        self.libtool_install = None

    def finalize_options(self):
        _install_lib.finalize_options(self)

        if self.libtool_install is None:
            self.libtool_install = 'libtool --mode=install install'
        self.cmd_libtool_install = split_quoted(self.libtool_install)

    def copy_tree(self, infile, outfile, preserve_mode=1, preserve_times=1,
                   preserve_symlinks=0, level=1):
        """Copy the build directory tree, respecting dry-run and force flags.
        Special treatment of libtool files.
        """
        if not self.dry_run and not os.path.isdir(infile):
            raise DistutilsFileError(
                  "cannot copy tree '%s': not a directory" % infile)
        try:
            names = os.listdir(infile)
        except OSError as e:
            if self.dry_run:
                names = []
            else:
                raise DistutilsFileError(
                      "error listing files in '%s': %s" % (infile, e.strerror))

        if not self.dry_run:
            mkpath(outfile)

        outputs = []

        for n in names:
            src_name = os.path.join(infile, n)
            dst_name = os.path.join(outfile, n)

            if n.startswith('.nfs'):
                # skip NFS rename files
                continue
            if n in ('.libs', '_libs'):
                # skip libtool directories
                continue

            if preserve_symlinks and os.path.islink(src_name):
                link_dest = os.readlink(src_name)
                log.info("linking %s -> %s", dst_name, link_dest)
                if not self.dry_run:
                    os.symlink(link_dest, dst_name)
                outputs.append(dst_name)

            elif os.path.isdir(src_name):
                outputs.extend(
                    self.copy_tree(src_name, dst_name, preserve_mode,
                                   preserve_times, preserve_symlinks))

            elif n.endswith('.la'):
                spawn(self.cmd_libtool_install + [ src_name, dst_name ],
                      dry_run=self.dry_run)

            else:
                copy_file(src_name, dst_name, preserve_mode,
                          preserve_times, not self.force,
                          dry_run=self.dry_run)
                outputs.append(dst_name)

        return outputs

def sysconfig_replace_ext(varname, newext):
    """Replace the extension in a system config variable.
    """
    from distutils.sysconfig import _config_vars
    val = _config_vars.get(varname, '')
    suffpos = val.rfind('.')
    if suffpos >= 0:
        val = val[:suffpos]
    _config_vars[varname] = val + newext

# change default shared object suffix to .la
from distutils.sysconfig import get_config_vars
get_config_vars()
sysconfig_replace_ext('EXT_SUFFIX', '.la')
sysconfig_replace_ext('SO', '.la')



#!/usr/bin/env python
# vim:sw=4 ts=4 et

import _addrxlat
from _addrxlat import BaseException

_exceptions = (
    ('NotImplementedError', _addrxlat.ERR_NOTIMPL, (NotImplementedError,)),
    ('NotPresentError', _addrxlat.ERR_NOTPRESENT),
    ('InvalidError', _addrxlat.ERR_INVALID),
    ('MemoryError', _addrxlat.ERR_NOMEM, (MemoryError,)),
    ('NoDataError', _addrxlat.ERR_NODATA),
    ('NoMethodError', _addrxlat.ERR_NOMETH),
)

def new_exception(name, status, addbases=()):
    '''Create an addrxlat exception for a given status code.
    The new exception is derived from BaseException, but you may specify
    additional base classes with addbases.
    '''

    def __init__(self, *args, **kwargs):
        super(cls, self).__init__(status, *args, **kwargs)

    def __repr__(self):
        "x.__repr__() <==> repr(x)"
        return "%s%r" % (self.__class__.__name__, self.args[1:])

    di = {
        '__doc__' : name + '([message])' + '''

        If message is not specified, use the default error message.
        ''',
        'status' : status,
        '__init__' : __init__,
        '__repr__' : __repr__,
    }

    cls = type(name, (BaseException,) + addbases, di)
    return cls

def get_exception(status, *args, **kwargs):
    '''get_exception(status[, message])

    Get an appropriate exception for the given status. If there is no
    specific exception, make an instance of BaseException.
    '''
    for exc in BaseException.__subclasses__():
        if status == exc.status:
            return exc(*args, **kwargs)
    return BaseException(status, *args, **kwargs)

for _exc in _exceptions:
    _cls = new_exception(*_exc)
    globals()[_cls.__name__] = _cls

# Free up temporary variables
del _exc, _cls, _exceptions
